unit RootCAModule;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  MPX509, ASN1, MPDL;

type
  TdmRootCA = class(TDataModule)
    X509TrustedCertificates1: TX509TrustedCertificates;
    X509Certificate1: TX509Certificate;
    X509TrustedCertificates2: TX509TrustedCertificates;
    RootCACert: TX509Certificate;
    procedure DataModuleDestroy(Sender: TObject);
  private
    FCAPriv: TDLPrivateKey;
    FCAPubl: TDLPublicKey;
  public
    procedure CreateKeyAndCert;
    procedure AddCRL(AStream: TStream);
    function HandleCertRequest(const CertReq: TASN1Struct;
                               OutputStream: TStream): Boolean;
    function VerifyCertificate(AStream: TStream;
                               var Status: TCertStatusCode): Boolean;
    property RootPublicKey: TDLPublicKey read FCAPubl;
  end;

var
  dmRootCA: TdmRootCA;

implementation

uses
  PKIX, X509, PKCS10;

{$R *.DFM}

{ TdmRootCA }

function TdmRootCA.HandleCertRequest(const CertReq: TASN1Struct;
  OutputStream: TStream): Boolean;
var
  Cert: TASN1Struct;
  Validity: TX509Validity;
  Status: TCertStatusCode;
begin
  Result := CheckSignatureCR(CertReq);
  if Result then begin
    Cert := nil;
    try
      Validity.notBefore := Now;
      Validity.notAfter := Now + 365;
      ComposeCertificateCR(Cert,CertReq,RootCACert.Cert,Validity);
      DSASignCertificate(Cert,RootCACert.Cert,FCAPriv);
      X509TrustedCertificates2.AddCertificate(Cert,False,Status);
      Cert.SaveToStream(OutputStream,fmtDER);
    finally
      Cert.Free;
    end;
  end;
end;

procedure TdmRootCA.DataModuleDestroy(Sender: TObject);
begin
  DisposeDLPrivateKey(FCAPriv);
  DisposeDLPublicKey(FCAPubl);
end;

procedure TdmRootCA.CreateKeyAndCert;  
var
  PK: TX509PublicKey;   
  Status: TCertStatusCode;
begin   
  if FCAPriv.X = nil then
    DLKeys(1024,160,FCAPriv,FCAPubl)
  else
    DLKeys(FCAPriv.Params,FCAPriv.X,FCAPubl.Y);
  PK := TX509PublicKey.Create;
  try
    TranslateDSAPublicKey(FCAPubl,PK);
    RootCACert.ImposePublicKey(PK);
  finally
    PK.Free;
  end;
  RootCACert.Subject.commonName := 'ROOT CA';
  RootCACert.KeyUsage := [cRLSign,keyCertSign];
  RootCACert.Validity.notBefore := Now;        
  RootCACert.Validity.notAfter := Now + 365;
  RootCACert.ComposeCACertificate(nil,1);
  RootCACert.DSASignCACertificate(FCAPriv);

  X509TrustedCertificates1.SignatureCert := RootCACert.Cert; 
  X509TrustedCertificates2.SignatureCert := RootCACert.Cert;
  X509TrustedCertificates1.AddCertificate(RootCACert.Cert,True,Status);
end;

function TdmRootCA.VerifyCertificate(AStream: TStream;
                                     var Status: TCertStatusCode): Boolean;
begin
  X509Certificate1.LoadFromStream(AStream,fmtDER);
  Result := X509Certificate1.VerifyCertificate(Status);
end;

procedure TdmRootCA.AddCRL(AStream: TStream);
var
  CRL: TASN1Struct;
  Status: TCertStatusCode;
begin
  CRL := TASN1Struct.Create;
  try
    CRL.LoadFromStream(AStream,fmtDER);
    X509TrustedCertificates2.AddCRL(CRL,Status);
    if Status <> crcOK then
      raise Exception.Create('CRL not accepted');
  finally
    CRL.Free;
  end;
end;

end.
